package node

import (
	"sort"

	pb "git.autistici.org/ale/autoradio/proto"
)

// mountStatus reports the configuration and status of a mount,
// including eventual transcoded mounts that source it.
type mountStatus struct {
	Mount       *pb.Mount
	Listeners   int
	TransMounts []*mountStatus
}

func newMountStatus(m *pb.Mount, nodes []*nodeInfo) *mountStatus {
	var listeners int
	for _, n := range nodes {
		for _, ims := range n.status.IcecastMounts {
			if ims.Path == m.Path {
				listeners += int(ims.Listeners)
				break
			}
		}
	}
	return &mountStatus{
		Mount:     m,
		Listeners: listeners,
	}
}

type mountStatusList []*mountStatus

func (l mountStatusList) Len() int      { return len(l) }
func (l mountStatusList) Swap(i, j int) { l[i], l[j] = l[j], l[i] }
func (l mountStatusList) Less(i, j int) bool {
	return l[i].Mount.Path < l[j].Mount.Path
}

// mountsToStatus converts a list of mounts (and eventually the
// current list of nodes) to a nicely sorted and tree-aggregated list
// of mountStatus objects. The list of nodes can be nil, in which case
// listener statistics will be omitted.
func mountsToStatus(mounts []*pb.Mount, nodes []*nodeInfo) []*mountStatus {
	// Aggregate stats, and create a tree of transcoding mounts.
	ms := make(map[string]*mountStatus)
	for _, m := range mounts {
		if m.HasTranscoder() {
			continue
		}
		ms[m.Path] = newMountStatus(m, nodes)
	}
	for _, m := range mounts {
		if !m.HasTranscoder() {
			continue
		}
		src := ms[m.TranscodeParams.SourcePath]
		if src == nil {
			continue
		}
		src.TransMounts = append(src.TransMounts, newMountStatus(m, nodes))
	}
	msl := make([]*mountStatus, 0, len(ms))
	for _, m := range ms {
		msl = append(msl, m)
	}

	// Sort everything (including transcoded mounts).
	sort.Sort(mountStatusList(msl))
	for _, s := range msl {
		if s.TransMounts != nil {
			sort.Sort(mountStatusList(s.TransMounts))
		}
	}
	return msl
}
