// Variant of the base http/httputil ReverseProxy suitable for
// low-latency, long-term connections.

// Copyright 2011 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// HTTP reverse proxy handler

package node

import (
	"io"
	"log"
	"net"
	"net/http"
	"net/url"
	"strings"
	"sync"
	"time"

	"github.com/prometheus/client_golang/prometheus"
)

func copyHeader(dst, src http.Header) {
	for k, vv := range src {
		for _, v := range vv {
			dst.Add(k, v)
		}
	}
}

// Hop-by-hop headers. These are removed when sent to the backend.
// http://www.w3.org/Protocols/rfc2616/rfc2616-sec13.html
var hopHeaders = []string{
	"Connection",
	"Keep-Alive",
	"Proxy-Authenticate",
	"Proxy-Authorization",
	"Te", // canonicalized version of "TE"
	"Trailers",
	"Transfer-Encoding",
	"Upgrade",
}

type wrappedWriter interface {
	WrappedWriter() http.ResponseWriter
}

// Proxy a request to the desired backend. Due to the way the Icecast
// protocol works, this just dumps the initial (rewritten) HTTP/1.0
// request, and then switches to a full bi-directional TCP proxy.  The
// outbound request is built from the target host, path, and eventual
// query string parameters and headers passed on from the original
// request. The additional streamName parameter is used for
// instrumentation.
func doIcecastProxy(rw http.ResponseWriter, req *http.Request, target *url.URL, streamName string) {
	outreq := new(http.Request)
	*outreq = *req // includes shallow copies of maps, but okay

	// Make a HTTP/1.0 connection to the backend.
	outreq.URL.Scheme = target.Scheme
	outreq.URL.Host = target.Host
	//outreq.URL.Path = singleJoiningSlash(target.Path, req.URL.Path)
	outreq.URL.Path = target.Path
	outreq.Proto = "HTTP/1.0"
	outreq.ProtoMajor = 1
	outreq.ProtoMinor = 0
	outreq.Close = true

	// Remove hop-by-hop headers to the backend. This is modifying
	// the same underlying map from req (shallow copied above) so
	// we only copy it if necessary.
	copiedHeaders := false
	for _, h := range hopHeaders {
		if outreq.Header.Get(h) != "" {
			if !copiedHeaders {
				outreq.Header = make(http.Header)
				copyHeader(outreq.Header, req.Header)
				copiedHeaders = true
			}
			outreq.Header.Del(h)
		}
	}

	if clientIP, _, err := net.SplitHostPort(req.RemoteAddr); err == nil {
		// If we aren't the first proxy retain prior
		// X-Forwarded-For information as a comma+space
		// separated list and fold multiple headers into one.
		if prior, ok := outreq.Header["X-Forwarded-For"]; ok {
			clientIP = strings.Join(prior, ", ") + ", " + clientIP
		}
		outreq.Header.Set("X-Forwarded-For", clientIP)
	}

	// Create the upstream connection and write the HTTP request
	// to it.
	upstream, err := net.Dial("tcp", outreq.URL.Host)
	if err != nil {
		log.Printf("http: proxy dial error: %v", err)
		rw.WriteHeader(http.StatusInternalServerError)
		return
	}
	defer upstream.Close()
	if err := outreq.Write(upstream); err != nil {
		log.Printf("http: proxy request write error: %v", err)
		rw.WriteHeader(http.StatusInternalServerError)
		return
	}

	// Hijack the request connection. We might need to unroll the
	// layers of nested WrappedWriters, until we find a
	// ResponseWriter that also implements the http.Hijacker
	// interface.
	var conn net.Conn
	for {
		if h, ok := rw.(http.Hijacker); ok {
			var err error
			conn, _, err = h.Hijack()
			if err != nil {
				log.Printf("http: proxy hijack error: %v", err)
				rw.WriteHeader(http.StatusInternalServerError)
				return
			}
			break
		} else if w, ok := rw.(wrappedWriter); ok {
			rw = w.WrappedWriter()
		} else {
			break
		}
	}
	if conn == nil {
		log.Printf("http: proxy error: could not find hijackable connection")
		rw.WriteHeader(http.StatusInternalServerError)
		return
	}
	defer conn.Close()
	if err := conn.SetDeadline(time.Time{}); err != nil {
		log.Printf("http: proxy setdeadline error: %v", err)
	}

	// Run two-way proxying.
	handleProxy(conn.(*net.TCPConn), upstream.(*net.TCPConn), streamName)
}

func copyStream(out, in *net.TCPConn, promCounter prometheus.Counter) {
	buf := getBuf()
	defer releaseBuf(buf)
	defer in.CloseRead()   //nolint
	defer out.CloseWrite() //nolint

	for {
		n, err := io.CopyBuffer(out, in, buf)
		promCounter.Add(float64(n))
		if err != nil {
			if err != io.EOF {
				log.Printf("http: proxy error: %v", err)
			}
			return
		}
	}
}

// Simple two-way TCP proxy that copies data in both directions and
// can shutdown each direction of the connection independently.
func handleProxy(conn *net.TCPConn, upstream *net.TCPConn, streamName string) {
	var wg sync.WaitGroup
	wg.Add(2)
	streamListeners.WithLabelValues(streamName).Inc()

	go func() {
		copyStream(conn, upstream, streamSentBytes.WithLabelValues(streamName))
		wg.Done()
	}()
	go func() {
		copyStream(upstream, conn, streamRcvdBytes.WithLabelValues(streamName))
		wg.Done()
	}()

	wg.Wait()
	streamListeners.WithLabelValues(streamName).Dec()
}

// Implementation of a simple buffer cache, to minimize large
// allocations at runtime.
const (
	bufSize     = 8192
	bufPoolSize = 512
)

var bufPool chan []byte

func init() {
	bufPool = make(chan []byte, bufPoolSize)
	for i := 0; i < bufPoolSize; i++ {
		bufPool <- make([]byte, bufSize)
	}
}

func getBuf() (b []byte) {
	select {
	case b = <-bufPool:
	default:
		b = make([]byte, bufSize)
	}
	return
}

func releaseBuf(b []byte) {
	select {
	case bufPool <- b:
	default:
	}
}
